// $RCSfile: MPClient.java,v $
// $Author: pmwilson $ 
// $Date: 2004/03/23 04:36:39 $
// $Revision: 1.38 $

package com.spamnbeans.client;

import com.spamnbeans.msg.MsgStats;
import com.spamnbeans.server.*;

import javax.naming.InitialContext;
import javax.naming.Context;
import javax.rmi.PortableRemoteObject;
import java.rmi.RemoteException;

import java.util.logging.*;
import java.io.*;
import java.net.*;

/**
 * Spam-n-Beans mail processing system client interface.
 * <p>
 * This client interface connects to a Mail Processing Server and submits an
 * email for processing.  This class is thread safe, as it maintains no
 * internal state.
 * <p>
 * If run directly, the email is read from standard input, and a copy of
 * the email message is printed on standard output along with any relevant
 * changes.  It will attempt to retry several times if errors occur.
 *
 * @author Andrew Boyer,
 *			<a href="mailto:aboyer@andrew.cmu.edu">aboyer@andrew.cmu.edu</a>
 * @version 1.0
 *
 * @author Philippe M. Wilson
 *			&lt;<a href="mailto:pmwilson@cmu.edu">pmwilson@cmu.edu</a>&gt;
 * @version 1.2
 *
 */
public class MPClient
{
    private static Logger logger = Logger.getLogger(MPClient.class.getName());

    private final ManagerRemote mgr;
    private static Integer clientID;
    private static String clientName;
    private static String clientIP;

	/**
	 * Create a new MPClient with automatic client ID
	 */
	public MPClient() throws ServerException {
		logger.info("Creating MPClient...");
		setupClientInfo();
		mgr = connectLB();
		authClient();
	}

	
	/**
	 * Convert an InetAddress into an ID.
	 *
	 * @param addr the address to convert
	 * @return the ID of the host given
	 */
	public static Integer getClientID(InetAddress addr){
		Integer clientID = new Integer(addr.hashCode());

		if (logger.isLoggable(Level.FINEST)) {
			logger.log(Level.FINEST, "Client hostname:\t"+addr.getHostName());
			logger.log(Level.FINEST, "Client ip:\t\t"+addr.getHostAddress());
		}
		
		logger.severe("ClientID:\t\t"+clientID.toString());
		return clientID;
	}


	/**
	 * Convert a string into an ID.
	 *
	 * @param id the string to convert (attempted as a hostname, then an integer)
	 * @return the ID of the host given
	 */
	public static Integer getClientID(String id){
		try {
			return getClientID(InetAddress.getByName(id));
		} catch (UnknownHostException e){
			try {
				return new Integer(id);
			} catch (NumberFormatException ne){
				return new Integer(0);
			}
		}
	}


	/**
	 * Get the ID of the current machine.
	 *
	 * @return the ID of the localhost
	 */
	public static Integer getClientID(){
		try {
			return getClientID(InetAddress.getLocalHost());
		} catch (UnknownHostException e){
			return new Integer(0);
		}
	}


	/**
	 * Store the ID (hashcode), IP, and hostname of the current machine.
	 *
	 */
	public static void setupClientInfo(){
	    try {
		InetAddress addr = InetAddress.getLocalHost();

		clientID = new Integer(addr.hashCode());
		clientName = addr.getHostName();
		clientIP = addr.getHostAddress();

		if (logger.isLoggable(Level.FINEST)) {
		    logger.log(Level.FINEST, "Client hostname:\t"+clientName);
		    logger.log(Level.FINEST, "Client ip:\t\t"+clientIP);
		}
	    
		logger.severe("ClientID:\t\t"+clientID.toString());
	    } catch (UnknownHostException e){
		clientID = new Integer(0);
		clientName = "default";
	    }
	}


	/**
	 * Connect to a Load Balancer.
	 */
	private ManagerRemote connectLB() throws ServerException{
		while(true){
			logger.fine("Connecting to Naming Service...");
			try {  // Set up a connection to the golden name server
				Context jndiContext = new InitialContext();

				Object mgrref = jndiContext.lookup("ManagerHomeRemote");
				logger.fine("Connected to Naming Service (JNDI)");

				ManagerHomeRemote mgrhome = (ManagerHomeRemote)
					PortableRemoteObject.narrow(mgrref,ManagerHomeRemote.class);

				ManagerRemote mgr = mgrhome.create();
				logger.fine("Connected to Load Balancer");
				return mgr;
			}
			catch (Throwable t) {
				String err = "Error connecting to Naming Service or Load Balancer";
				if (logger.isLoggable(Level.FINEST)) {
					logger.log(Level.FINEST, err, t);
				}
				else {
					logger.severe(err);
				}

				throw new ServerException(err, t);
			}

			sleep(1000 * 60); // wait 60 seconds before trying again
		}
	}


	/**
	 * Authenticate the clientID stored in this class.
	 * Updated to register us, not see if we are already registered.
	 */
	private void authClient() throws ServerException {	
		logger.info("Authenticating client...");
		try {
		    mgr.clientCreate(clientID, clientName);
			/*if (!mgr.clientExists(clientID)) {
			  String err = "INVALID ClientID: "+clientID.toString();
			  logger.severe(err);
			  throw new InvalidClientException(err);
			}*/
		} catch (DuplicateClientException dce) {
		    // Ignore
		    String err = "Authenticated duplicate clientID: "+clientID.toString();
		    logger.info(err);
		} catch (Throwable t) {
		    String err = "Error authenticating clientID: "+clientID.toString();
		    logger.info(err);
		    throw new ServerException(err, t);
		}

	}


	/**
	 * Connects to a Processing Server and submits a message for processing.
	 *
	 * @param msg the message to process
	 * @return the headers generated by the Processing Server
	 */
	public MsgStats processMessage(String msg)
		throws InvalidClientException, InvalidServerException, ServerException, RemoteException {
		String mpname = null; // name of MailProc server

		if (msg==null || msg.length() == 0) {
			return new MsgStats();
		}

		Integer msgID = new Integer((int)System.currentTimeMillis());

		StringBuffer mpName = new StringBuffer();

		while (true) {
			try {
				MailProcRemote mp = connectToServer(mpName);
				mpname = mpName.toString();

				// Check the message
				return mp.processMessage(clientID, msgID, msg);
			}
			catch (ServerException se) { // non-fatal failover
			    mgr.serverSetEnabled(mpName.toString(), false);
			}
			// TODO: does this actually make sense based on the conditions 
			// where ProcessingFailureException is thrown by MailProcBean ?
			// should we really be disabling the bean?  most of the conditions 
			// happen when the client sends a bad request
			catch (ProcessingFailureException pfe) { // non-fatal failover
				String err;
				if(mpname==null)
					err = "Error processing message on unknown server";
				else
					err = "Error processing message on: " + mpname;
				if (logger.isLoggable(Level.FINEST)) {
					logger.log(Level.FINEST, err, pfe);
				} else {
					logger.severe(err);
				}
				mgr.serverSetEnabled(mpname.toString(), false);
			}
		}
	}


	/**
	 * Contacts the Load Balancer to find a Processing Server and initiates a
	 * connection.
	 *
	 * @return the remote interface of a ready/waiting Processing Server
	 */
	private MailProcRemote connectToServer() throws ServerException {
		return connectToServer((StringBuffer)null);
	}

	
	/**
	 * Contacts the Load Balancer to find a Processing Server and initiates a
	 * connection.
	 *
	 * @param mpName a StringBuffer to fill with the name of the server matching the remote interface
	 * @return the remote interface of a ready/waiting Processing Server
	 */
	private MailProcRemote connectToServer(StringBuffer mpName) throws ServerException {
		logger.fine("Connecting to Processing Server");
		try {
			if(mpName!=null) mpName.setLength(0);
			String mpname = mgr.serverLoadBalance();
			if ( mpname == null ){
				String err = "Load Balancer returned no server(s)";
				logger.severe(err);
				throw new ServerException(err);
			}
			if(mpName!=null) mpName.append(mpname);

			logger.fine("Directing request to "+mpname+" ...");
			
			Context jndiContext = new InitialContext();
			jndiContext.addToEnvironment("java.naming.provider.url", mpname);
			logger.fine("Connected to Naming Service (JNDI)");
			
			Object mpref = jndiContext.lookup("MailProcHomeRemote");

			MailProcHomeRemote mphome = (MailProcHomeRemote)
				PortableRemoteObject.narrow(mpref, MailProcHomeRemote.class);

			MailProcRemote mp = mphome.create();
			mp.attachServer(mpname);
			logger.fine("Attached to Processing Server: "+mp.getServerName());
			
			return mp;
		}
		catch (ServerException se){
			String err = "Error connecting to Processing Server";
			if (logger.isLoggable(Level.FINEST)) {
				logger.log(Level.FINEST, err, se);
			} else {
				logger.warning(err);
			}

			throw new ServerException(err, se);
		}
		catch (Throwable t) {
			//TODO: Could this ever be non-fatal?
			//      yes, if:
			//			retry X times
			//			mgr.serverLoadBalance() returns a list
			//			have a host/list to try before needing to do a lookup
			//				try connecting to last server used
			//				load/store a list of servers from a file
			String err = "Error connecting to Processing Server";
			if (logger.isLoggable(Level.FINEST)) {
				logger.log(Level.FINEST, err, t);
			} else {
				logger.warning(err);
			}

			throw new ServerException(err, t);
		}
	}


	/**
	 * Reads a message from an InputStream and separates it into two strings
	 * divided just before the first blank line.
	 * <p>
	 * If there is no blank line in the input, the second string will be null.
	 *
	 * @param in the stream to use when reading the message
	 * @return the message in two strings (split before first blank line)
	 */
	public static String[] readMessage(InputStream in) throws IOException {
		String header = null;
		String body = null;
		BufferedReader br = new BufferedReader(new InputStreamReader(in));
		StringBuffer buf = new StringBuffer();
		String inputLine;

		logger.finer("Getting message...");

		while ((inputLine = br.readLine()) != null) {
			if (header == null && inputLine.length() == 0) {
				header = buf.toString();
				buf = new StringBuffer();
			}
			buf.append(inputLine);
			buf.append("\r\n");
		}

		body = buf.toString();

		br.close();

		if (header == null) {
			// pretend like entire message is header
			header = body;
			body = "";
		}

		logger.finer("Message parsed");
		return new String[] {header, body};
	}


	public static void main(String [] args) throws Throwable
	{
		String[] s = null;

		logger.info("Running end-to-end use case...");

		logger.info("Getting message...");
		try{
			s = readMessage(System.in);
			if(s==null)
				throw new IOException(); // act like a read failure occurred
		}catch(IOException e){
			if (logger.isLoggable(Level.FINEST)) {
				logger.log(Level.FINEST, "Error reading input message:", e);
			}else {
				logger.severe("Failed to read input message");
			}
			System.exit(100);
		}

		String header = s[0];
		String body = s[1];
		
		//TODO: wrap w/ throw/catch blocks
		MPClient client = new MPClient();
		MsgStats stats = client.processMessage(header + body);
		String newHeaders = stats.getNewHeaders();
		String newMessage = header + newHeaders + body;

		// a futile effort to normalize line endings; alas
		String[] lines = newMessage.split("\n");
		for (int i = 0; i < lines.length; i++) {
			System.out.println(lines[i]);
		}
	}
}
