// $Id: MailProcBean.java,v 1.20 2004/03/21 21:45:38 cfry Exp $

package com.spamnbeans.server;

import com.spamnbeans.spam.*;
import com.spamnbeans.msg.*;

import javax.rmi.PortableRemoteObject;
import java.rmi.RemoteException;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.ejb.SessionBean;
import javax.ejb.EJBException;
import javax.ejb.FinderException;
import javax.ejb.CreateException;
import java.io.IOException;

/**
 * Define the MailProc SessionBean methods.
 * 
 * <p>
 * The MailProc Bean gets requests from clients, interacts with ServerStat,
 * ClientStat, and MsgStat EntityBeans, and processes messages using SpamC
 * SessionBeans.
 * 
 * @author Andrew Boyer, <a href="mailto:aboyer@andrew.cmu.edu">
 *         aboyer@andrew.cmu.edu</a>
 * @version 1.0
 */
public class MailProcBean implements SessionBean {
	private Context jndiContext;
	private ServerStatLocal ssl;

	/**
	 * Create a new MailProcBean
	 */
	public void ejbCreate() throws RemoteException {
	}

	/**
	 * Attach a server to the MailProcBean Gets and stores initial HomeLocal
	 * interface for Server
	 */
	public void attachServer(String name) throws RemoteException {
		try {
			jndiContext = new InitialContext();
			Object obj = jndiContext.lookup("ServerStatHomeLocal");

			ServerStatHomeLocal home =
				(ServerStatHomeLocal) PortableRemoteObject.narrow(
					obj,
					ServerStatHomeLocal.class);

			ssl = home.findByName(name);
		} catch (Exception e) {
			throw new EJBException(e);
		}
	}

	/**
	 * Return the ServerID of the currently connected Server
	 */
	public Integer getServerID() throws RemoteException {
		try {
			return ssl.getServerID();
		} catch (Exception e) {
			throw new EJBException(e);
		}
	}

	/**
	 * Return the Name of the currently connected Server
	 */
	public String getServerName() throws RemoteException {
		try {
			return ssl.getName();
		} catch (Exception e) {
			throw new EJBException(e);
		}
	}

	/**
	 * Process a single e-mail message
	 * 
	 * @param clientID
	 *            The ClientID of the requesting Client
	 * @param message
	 *            The full text of the e-mail to check
	 * @return The extra e-mail headers generated
	 */
	public MsgStats processMessage(
		Integer clientID,
		Integer msgID,
		String message)
		throws RemoteException, InvalidClientException, ProcessingFailureException {
		MsgStats stats = new MsgStats();
		long serverStartTime = System.currentTimeMillis();
		SpamCRemote spamcb;
		MsgLocal msg;
		ClientStatLocal csl;

		// Validate message - anything else to check?
		if (message == null || message.length() == 0)
			throw new ProcessingFailureException("No message specified");

		// Validate ClientID argument with ClientStat entity bean
		try {
			Object obj = jndiContext.lookup("ClientStatHomeLocal");

			ClientStatHomeLocal home =
				(ClientStatHomeLocal) PortableRemoteObject.narrow(
					obj, ClientStatHomeLocal.class);

			csl = home.findByPrimaryKey(clientID);
		} catch (FinderException fe) {
			throw new InvalidClientException(
				"Client not registered: " + clientID);
		} catch (Throwable t) {
			throw new InvalidClientException("Client database not found");
		}

		try { // Make a database entry / Msg Entity Bean
			Object obj = jndiContext.lookup("MsgHomeLocal");
			MsgHomeLocal msghome = (MsgHomeLocal) obj;

			try { // Make a new one
				long dbStartTime = System.currentTimeMillis();
				msg =
					msghome.createMsg(
						msgID,
						clientID,
						getServerID(),
						"Not Implemented",
						new Integer(message.length()),
						new Integer(0));
				long dbEndTime = System.currentTimeMillis();
				stats.setDbTime(dbEndTime - dbStartTime);
			} catch (CreateException ce) { // Try to reuse the old one
				msg = msghome.findByPrimaryKey(msgID);
				msg.setClientID(clientID);
				msg.setServerID(getServerID());
				msg.setTimeStamp("Not Implemented");
				msg.setMsgSize(new Integer(message.length()));
				msg.setMsgState(new Integer(0));
			}

		} catch (Throwable t) {
			throw new ProcessingFailureException("Couldn't load MsgBean");
		}

		try { // Make a SpamC Session Bean
			Object ref = jndiContext.lookup("SpamCHomeRemote");
			SpamCHomeRemote home =
				(SpamCHomeRemote) PortableRemoteObject.narrow(
					ref,
					SpamCHomeRemote.class);

			spamcb = home.create();
		} catch (Throwable t) {
			throw new ProcessingFailureException("Couldn't load SpamCBean");
		}

		try { // Check the message
			long spamStartTime = System.currentTimeMillis();
			String hdrs = spamcb.processMessage(message);
			long spamEndTime = System.currentTimeMillis();
			stats.setSpamTime(spamEndTime - spamStartTime);
			stats.setNewHeaders(hdrs);
		} catch (IOException ie) {
			throw new ProcessingFailureException("Communication error", ie);
		} catch (SpamException e) {
			throw new ProcessingFailureException("Spam error", e);
		}

		try { // Update the Msg Bean
			msg.setMsgState(new Integer(1));
		} catch (Throwable t) {
			throw new ProcessingFailureException("Couldn't update MsgBean");
		}

		/*
		 * try { // DEBUG - delete the Msg Bean msg.remove(); } catch
		 * (Exception e) { throw new ProcessingFailureException("Couldn't
		 * delete MsgBean");
		 */

		// Update the ReplicaState with the number of requests processed
		try {
			Object obj = jndiContext.lookup("ReplicaStateHomeLocal");
			ReplicaStateHomeLocal home =
				(ReplicaStateHomeLocal) PortableRemoteObject.narrow(
						obj,
						ReplicaStateHomeLocal.class);
			ReplicaStateLocal rsl = home.create();
			rsl.addRequest();
		} catch (Throwable t) {
			throw new ProcessingFailureException("Unable to update ReplicaState");
		}

		long serverEndTime = System.currentTimeMillis();
		stats.setServerTime(serverEndTime - serverStartTime);
		return stats;
	}

	/**
	 * Not implemented
	 */
	public void ejbRemove() {
	}

	/**
	 * Not implemented
	 */
	public void ejbActivate() {
	}

	/**
	 * Not implemented
	 */
	public void ejbPassivate() {
	}

	/**
	 * Not implemented
	 */
	public void setSessionContext(javax.ejb.SessionContext cntx) {
	}
}
